from dataclasses import dataclass
from pathlib import Path
from typing import Annotated, Dict, List

from pvi._schema_utils import as_discriminated_union, desc
from pvi.device import Device, DeviceRef


@dataclass
class IndexEntry:
    label: Annotated[str, desc("Button label")]
    file_name: Annotated[str, desc("File name of UI file")]
    macros: Annotated[Dict[str, str], desc("Macros to launch UI with")]


@as_discriminated_union
@dataclass
class Formatter:
    def format(self, device: Device, prefix: str, path: Path):
        raise NotImplementedError(self)

    def format_index(self, label: str, index_entries: List[IndexEntry], path: Path):
        """Format an index of buttons to open the given UIs.

        Args:
            label: Title of generated index UI
            index_entries: Buttons to format on the index UI
            path: Output path of generated UI

        """
        self.format(
            Device(
                label,
                children=[
                    DeviceRef(
                        index.label,
                        pv=index.label.upper(),
                        ui=index.file_name,
                        macros=index.macros,
                    )
                    for index in index_entries
                ],
            ),
            prefix="Index",
            path=path,
        )



from dataclasses import dataclass, field
from enum import Enum
from typing import Annotated, ClassVar, Dict, List, Optional, Type, cast

from pvi._schema_utils import as_discriminated_union, desc, rec_subclasses
from pvi.device import (
    LED,
    CheckBox,
    ComboBox,
    Component,
    Named,
    ReadWidget,
    TextRead,
    TextWrite,
    WriteWidget,
)


@dataclass
class TypeStrings:
    """The type strings for record dtypes and parameter names"""

    asyn_read: Annotated[str, desc("e.g. asynInt32, asynOctetRead")]
    asyn_write: Annotated[str, desc("e.g. asynInt32, asynOctetWrite")]
    asyn_param: Annotated[str, desc("e.g. asynParamInt32, asynParamOctet")]


AReadWidget = Annotated[Optional[ReadWidget], desc("Widget to use for read record")]
AWriteWidget = Annotated[Optional[WriteWidget], desc("Widget to use for write record")]


class Access(Enum):
    """What access does the user have. One of:

    - R: Read only value that cannot be set. E.g. chipTemperature on a detector,
      or isHomed for a motor
    - W: Write only value that can be written to, but there is no current value.
      E.g. reboot on a detector, or overwriteCurrentPosition for a motor
    - RW: Read and Write value that can be written to and read back.
      E.g. acquireTime on a detector, or velocity of a motor
    """

    R = "R"  #: Read record only
    W = "W"  #: Write record only
    RW = "RW"  #: Read and write record

    def needs_read_record(self):
        return self != self.W

    def needs_write_record(self):
        return self != self.R


class DisplayForm(Enum):
    """Instructions for how a number should be formatted for display"""

    #: Use the default representation from value
    DEFAULT = "Default"
    #: Force string representation, most useful for array of bytes
    STRING = "String"
    #: Binary, precision determines number of binary digits
    BINARY = "Binary"
    #: Decimal, precision determines number of digits after decimal point
    DECIMAL = "Decimal"
    #: Hexadecimal, precision determines number of hex digits
    HEX = "Hex"
    #: Exponential, precision determines number of digits after decimal point
    EXPONENTIAL = "Exponential"
    #: Exponential where exponent is multiple of 3, precision determines number of
    #: digits after decimal point
    ENGINEERING = "Engineering"


@as_discriminated_union
@dataclass
class AsynParameter(Named):
    """Base class for all Asyn Parameters to inherit from"""

    type_strings: ClassVar[TypeStrings]
    read_record: Annotated[
        Optional[str], desc("The full read record, if not given then use $(name)_RBV")
    ] = None
    write_record: Annotated[
        Optional[str], desc("The full write record, if not given then use $(name)")
    ] = None
    display_form: Annotated[
        Optional[DisplayForm], desc("Display form for numeric/array fields")
    ] = None
    read_widget: AReadWidget = None
    write_widget: AWriteWidget = None

    def get_read_record(self) -> str:
        if self.read_record:
            return self.read_record
        else:
            return self.name + "_RBV"

    def get_write_record(self) -> str:
        if self.write_record:
            return self.write_record
        else:
            return self.name


@dataclass
class AsynBinary(AsynParameter):
    """Asyn Binary Parameter and records"""

    type_strings = TypeStrings(
        asyn_read="asynInt32",
        asyn_write="asynInt32",
        asyn_param="asynParamInt32",
    )
    read_widget: AReadWidget = field(default_factory=LED)
    write_widget: AWriteWidget = field(default_factory=CheckBox)


@dataclass
class AsynBusy(AsynBinary):
    """Asyn Busy Parameter and records"""


@dataclass
class AsynFloat64(AsynParameter):
    """Asyn Float64 Parameter and records"""

    type_strings: ClassVar[TypeStrings] = TypeStrings(
        asyn_read="asynFloat64",
        asyn_write="asynFloat64",
        asyn_param="asynParamFloat64",
    )
    read_widget: AReadWidget = field(default_factory=TextRead)
    write_widget: AWriteWidget = field(default_factory=TextWrite)


@dataclass
class AsynInt32(AsynParameter):
    """Asyn Int32 Parameter and records"""

    type_strings: ClassVar[TypeStrings] = TypeStrings(
        asyn_read="asynInt32",
        asyn_write="asynInt32",
        asyn_param="asynParamInt32",
    )
    read_widget: AReadWidget = field(default_factory=TextRead)
    write_widget: AWriteWidget = field(default_factory=TextWrite)


@dataclass
class AsynLong(AsynInt32):
    """Asyn Long Parameter and records"""


@dataclass
class AsynMultiBitBinary(AsynParameter):
    """Asyn MultiBitBinary Parameter and records"""

    type_strings: ClassVar[TypeStrings] = TypeStrings(
        asyn_read="asynInt32",
        asyn_write="asynInt32",
        asyn_param="asynParamInt32",
    )
    read_widget: AReadWidget = field(default_factory=TextRead)
    write_widget: AWriteWidget = field(default_factory=ComboBox)


@dataclass
class AsynString(AsynParameter):
    """Asyn String Parameter and records"""

    type_strings: ClassVar[TypeStrings] = TypeStrings(
        asyn_read="asynOctetRead",
        asyn_write="asynOctetWrite",
        asyn_param="asynParamOctet",
    )
    read_widget: AReadWidget = field(default_factory=TextRead)
    write_widget: AWriteWidget = field(default_factory=TextWrite)


InRecordTypes = dict(
    ai=AsynFloat64,
    bi=AsynBinary,
    longin=AsynLong,
    mbbi=AsynMultiBitBinary,
    stringin=AsynString,
)


OutRecordTypes = dict(
    ao=AsynFloat64,
    bo=AsynBinary,
    busy=AsynBusy,
    longout=AsynLong,
    mbbo=AsynMultiBitBinary,
    stringout=AsynString,
)


@dataclass
class AsynWaveform(AsynParameter):
    """Asyn Waveform Parameter and records"""

    type_strings: ClassVar[TypeStrings] = TypeStrings(
        asyn_read="asynOctetRead",
        asyn_write="asynOctetWrite",
        asyn_param="asynParamOctet",
    )
    read_widget: AReadWidget = field(default_factory=TextRead)
    write_widget: AWriteWidget = field(default_factory=TextWrite)


@dataclass
class AsynInt32Waveform(AsynWaveform):
    """Asyn Waveform Parameter and records with int32 array elements"""

    type_strings: ClassVar[TypeStrings] = TypeStrings(
        asyn_read="asynInt32ArrayIn",
        asyn_write="asynInt32ArrayOut",
        asyn_param="asynParamInt32",
    )


@dataclass
class AsynFloat64Waveform(AsynWaveform):
    """Asyn Waveform Parameter and records with int32 array elements"""

    type_strings: ClassVar[TypeStrings] = TypeStrings(
        asyn_read="asynFloat64ArrayIn",
        asyn_write="asynFloat64ArrayOut",
        asyn_param="asynParamFloat64",
    )


WaveformRecordTypes = [AsynWaveform] + cast(
    List[Type[AsynWaveform]], rec_subclasses(AsynWaveform)
)


def get_waveform_parameter(dtyp: str):
    for waveform_cls in WaveformRecordTypes:
        if dtyp in (
            waveform_cls.type_strings.asyn_read,
            waveform_cls.type_strings.asyn_write,
        ):
            return waveform_cls

    assert False, f"Waveform type for DTYP {dtyp} not found in {WaveformRecordTypes}"


@dataclass
class Record:
    name: str  # The name of the record e.g. $(P)$(M)Status
    type: str  # The record type string e.g. ao, stringin
    fields: Dict[str, str]  # The record fields
    infos: Dict[str, str]  # Any infos to be added to the record


class Parameter:
    invalid = ["DESC", "DTYP", "INP", "OUT", "PINI", "VAL"]

    def _remove_invalid(self, fields: Dict[str, str]) -> Dict[str, str]:
        valid_fields = {
            key: value for (key, value) in fields.items() if key not in self.invalid
        }
        return valid_fields

    def generate_component(self) -> Component:
        raise NotImplementedError(self)


class ReadParameterMixin:
    def _get_read_record(self) -> Optional[str]:
        raise NotImplementedError(self)



from copy import deepcopy
from typing import Dict, List, Optional, Sequence

from lxml.etree import ElementBase, SubElement, XMLParser, parse

from pvi._format.utils import Bounds
from pvi._format.widget import UITemplate, WidgetFormatter
from pvi.device import (
    LED,
    ComboBox,
    TableRead,
    TableWidgetType,
    TableWrite,
    TextFormat,
    TextRead,
    TextWrite,
    WidgetType,
    WriteWidget,
)

BOB_TEXT_FORMATS = {
    TextFormat.decimal: "1",
    TextFormat.hexadecimal: "4",
    TextFormat.engineer: "3",
    TextFormat.exponential: "2",
    TextFormat.string: "6",
}


class BobTemplate(UITemplate[ElementBase]):
    """Extract and modify elements from a template .bob file."""

    def __init__(self, text: str):
        """Parse an XML string to an element tree object."""

        # Passing `remove_blank_text` means we can pretty print our additions
        self.tree = parse(text, parser=XMLParser(remove_blank_text=True))
        self.screen = self.search("Display")

    def set(
        self,
        template: ElementBase,
        bounds: Optional[Bounds] = None,
        widget: Optional[WidgetType] = None,
        **properties,
    ) -> ElementBase:
        if bounds:
            properties["x"] = bounds.x
            properties["y"] = bounds.y
            properties["width"] = bounds.w
            properties["height"] = bounds.h

        widget_type = template.attrib.get("type", "")

        t_copy = deepcopy(template)
        for item, value in properties.items():
            new_text = ""

            match widget_type, item, value:
                case "table", "pv_name", pv:
                    new_text = f"pva://{pv}"  # Must include pva prefix
                case "action_button", "file", file_name:
                    new_text = file_name
                    if not new_text.endswith(".bob"):
                        new_text += ".bob"  # Must include file extension
                case "action_button", "macros", dict() as macros:
                    if macros:
                        add_button_macros(t_copy, value)
                case _:
                    new_text = str(value)

            if new_text:
                find_element(t_copy, item).text = new_text

        # Add additional properties from widget
        match widget_type, widget:
            case ("combo", ComboBox() as combo_box):
                add_combo_box_items(t_copy, combo_box)
            case ("table", TableRead() | TableWrite() as table):
                add_table_columns(t_copy, table)
            case (
                ("textentry", TextWrite(_, format))
                | ("textupdate", TextRead(_, format))
            ) if format is not None:
                add_format(t_copy, BOB_TEXT_FORMATS[format])

        return t_copy

    def search(self, search: str) -> ElementBase:
        """Locate and extract elements from the Element tree.

        Args:
            search: The unique name of the element to extract.
                Can be found in its name subelement.

        Returns:
            The extracted element.
        """

        tree_copy = deepcopy(self.tree)
        # 'name' is the unique ID for each element
        matches = [
            element.getparent()
            for element in tree_copy.iter("name")
            if element.text == search
        ]
        assert len(matches) == 1, f"Got {len(matches)} matches for {search!r}"

        # Isolate the screen properties
        if matches[0].tag == "display":
            for child in matches[0]:
                if child.tag == "widget":
                    matches[0].remove(child)

        return matches[0]

    def create_group(
        self,
        group_object: List[ElementBase],
        children: List[WidgetFormatter[ElementBase]],
        padding: Bounds = Bounds(),
    ) -> List[ElementBase]:
        """Create an xml group object from a list of child widgets

        Args:
            group_object: Templated group xml element
            children: List of child widgets within the group
            padding: Additional placement data to fit the children to the group object.
                Defaults to Bounds().

        Returns:
            An xml group with children attached as subelements
        """
        assert (
            len(group_object) == 1
        ), f"Size of group_object is {len(group_object)}, should be 1"
        for c in children:
            group_object[0].append(c.format()[0])
        return group_object


def add_table_columns(widget_element: ElementBase, table: TableWidgetType):
    if not table.widgets:
        # Default empty -> get options from pv
        return

    columns_element = SubElement(widget_element, "columns")
    for column, widget in enumerate(table.widgets):
        add_table_column(columns_element, f"Column {column + 1}", widget)

    add_editable(widget_element, editable=isinstance(table, TableWrite))


def add_table_column(
    columns_element: ElementBase,
    name: str,
    widget: WidgetType,
):
    options: Sequence[str] = []
    if isinstance(widget, LED):
        options = ["false", "true"]
    elif isinstance(widget, ComboBox):
        options = widget.choices

    column_element = SubElement(columns_element, "column")
    SubElement(column_element, "name").text = name
    add_editable(column_element, editable=isinstance(widget, WriteWidget))

    if options:
        options_element = SubElement(column_element, "options")
        for option in options:
            SubElement(options_element, "option").text = option


def add_button_macros(widget_element: ElementBase, macros: Dict[str, str]):
    """Add action macros to the given element.

    Args:
        widget_element: Element to add action macros to
        macros: Macros to add to element

    """
    action_element = find_element(widget_element, "action")
    macros_element = SubElement(action_element, "macros")
    for macro, value in macros.items():
        SubElement(macros_element, macro).text = value


def add_combo_box_items(widget_element: ElementBase, combo_box: ComboBox):
    if not combo_box.choices:
        # Default empty -> get items from pv
        return

    items_element = SubElement(widget_element, "items")
    for item in combo_box.choices:
        SubElement(items_element, "item").text = item

    SubElement(widget_element, "items_from_pv").text = "false"


def add_editable(element: ElementBase, editable: bool):
    SubElement(element, "editable").text = "true" if editable else "false"


def add_format(element: ElementBase, format: str):
    if format:
        SubElement(element, "format").text = format


def find_element(root_element: ElementBase, tag: str, index: int = 0) -> ElementBase:
    """Iterate tree to find tag and replace text.

    Args:
        root_element: Root of tree to search
        tag: Tag to search for in tree
        index: Match to return if multiple matches are found

    """
    match list(root_element.iter(tag=tag)):
        case []:
            raise ValueError(
                f"No matches for '{tag}' in {root_element.find('name').text}"
            )
        case matches:
            return matches[index]



from dataclasses import dataclass
from pathlib import Path
from typing import List

from jinja2 import Template

from pvi.device import Device, SignalR, SignalRW, SignalW, signal_access_mode, walk

PVI_TEMPLATE = Path(__file__).parent / "pvi.template.jinja"


@dataclass
class PviRecord:
    name: str
    access: str


def format_template(device: Device, prefix: str, output: Path):
    records: List[PviRecord] = []
    for node in walk(device.children):
        match node:
            case SignalR() | SignalW() | SignalRW() as signal:
                records.append(PviRecord(signal.pv, signal_access_mode(signal)))

    with output.open("w") as expanded, open(PVI_TEMPLATE, "r") as template:
        template_txt = Template(template.read()).render(
            records=records, prefix=prefix, device=device.label
        )
        expanded.write(template_txt)



from __future__ import annotations

import json
import re
from dataclasses import dataclass, field, fields
from enum import Enum
from pathlib import Path
from typing import (
    Annotated,
    Any,
    Callable,
    Dict,
    Generic,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    TypeVar,
    Union,
)

from apischema import deserialize, serialize
from ruamel.yaml import YAML

from pvi._schema_utils import add_type_field, as_discriminated_union, desc
from pvi.utils import find_pvi_yaml

PASCAL_CASE_REGEX = re.compile(r"(?<![A-Z])[A-Z]|[A-Z][a-z/d]|(?<=[a-z])\d")


def to_title_case(pascal_s: str) -> str:
    """Takes a PascalCaseFieldName and returns an Title Case Field Name

    Args:
        pascal_s: E.g. PascalCaseFieldName
    Returns:
        Title Case converted name. E.g. Pascal Case Field Name
    """
    return PASCAL_CASE_REGEX.sub(lambda m: " " + m.group(), pascal_s)[1:]


def to_snake_case(pascal_s: str) -> str:
    """Takes a PascalCaseFieldName and returns a snake_case_field_name
    Args:
        pascal_s: E.g. PascalCaseFieldName
    Returns:
        snake_case converted name. E.g. pascal_case_field_name
    """
    return PASCAL_CASE_REGEX.sub(lambda m: "_" + m.group().lower(), pascal_s)[1:]


class TextFormat(Enum):
    """Format to use for display of Text{Read,Write} widgets on a UI"""

    decimal = 0
    hexadecimal = 1
    engineer = 2
    exponential = 3
    string = 4


@as_discriminated_union
@dataclass
class ReadWidget:
    """Widget that displays a scalar PV"""


class LED(ReadWidget):
    """LED display of a boolean PV"""


@dataclass
class BitField(ReadWidget):
    """LED and label for each bit of an int PV"""

    labels: Annotated[Sequence[str], desc("Label for each bit")]


class ProgressBar(ReadWidget):
    """Progress bar from lower to upper limit of a float PV"""


@dataclass
class TextRead(ReadWidget):
    """Text view of any PV"""

    lines: Annotated[int, desc("Number of lines to display")] = 1
    format: Annotated[Optional[TextFormat], desc("Display format")] = None


@dataclass
class ArrayTrace(ReadWidget):
    """Trace of the array in a plot view"""

    axis: Annotated[
        str,
        desc(
            "Traces with same axis name will appear on same axis, "
            "plotted against 'x' trace if it exists, or indexes if not. "
            "Only one traces with axis='x' allowed."
        ),
    ]


@dataclass
class TableRead(ReadWidget):
    """Tabular view of an NTTable"""

    widgets: Annotated[
        Sequence[ReadWidget],
        desc("For each column, what widget should be repeated for every row"),
    ] = field(default_factory=list)


class ImageRead(ReadWidget):
    """2D Image view of an NTNDArray"""


@as_discriminated_union
@dataclass
class WriteWidget:
    """Widget that controls a PV"""


class CheckBox(WriteWidget):
    """Checkable control of a boolean PV"""


@dataclass
class ComboBox(WriteWidget):
    """Selection of an enum PV"""

    choices: Annotated[Sequence[str], desc("Choices to select from")] = field(
        default_factory=list
    )


@dataclass
class ButtonPanel(WriteWidget):
    """One-or-more buttons that poke a PV with a value

    Args:
        actions: Dict of button label to value the button sends

    """

    actions: Dict[str, Any] = field(default_factory=lambda: dict(go=1))


@dataclass
class TextWrite(WriteWidget):
    """Text control of any PV"""

    lines: Annotated[int, desc("Number of lines to display")] = 1
    format: Annotated[Optional[TextFormat], desc("Display format")] = None


@dataclass
class ArrayWrite(WriteWidget):
    """Control of an array PV"""

    widget: Annotated[WriteWidget, desc("What widget should be used for each item")]


@dataclass
class TableWrite(WriteWidget):
    widgets: Annotated[
        Sequence[WidgetType],
        desc("For each column, what widget should be repeated for every row"),
    ] = field(default_factory=list)


WidgetType = Union[ReadWidget, WriteWidget]
TableWidgetType = Union[TableRead, TableWrite]
TableWidgetTypes = (TableRead, TableWrite)


@as_discriminated_union
@dataclass
class Layout:
    """Widget displaying child Components"""


class Plot(Layout):
    """Children are traces of the plot"""


@dataclass
class Row(Layout):
    """Children are columns in the row"""

    header: Annotated[
        Optional[Sequence[str]],
        desc("Labels for the items in the row, None means use previous row header"),
    ] = None


@dataclass
class Grid(Layout):
    """Children are rows in the grid"""

    labelled: Annotated[bool, desc("If True use names of children as labels")] = True


@dataclass
class SubScreen(Layout):
    """Children are displayed on another screen opened with a button."""

    labelled: Annotated[bool, desc("Display labels for components")] = True


@dataclass
class Named:
    name: Annotated[
        str, desc("PascalCase name to uniquely identify", pattern=r"([A-Z][a-z0-9]*)*$")
    ]


class Labelled(Named):
    label: str

    def __init_subclass__(cls, **kwargs):
        # Add an optional label to the end of the dataclass if not already there
        has_label_field = [f for f in fields(cls) if f.name == "label"]
        # Hack so this doesn't fire for the Component class below, or anything
        # else that doesn't add annotations. We really want this to be on terminal
        # classes, but no way of knowing this in __init_subclass__
        adds_annotations = bool(cls.__annotations__)
        if not has_label_field and adds_annotations:
            cls.__annotations__["label"] = Annotated[
                str, desc("Label for GUI. If empty, use name in Title Case")
            ]
            cls.label = ""
            dataclass(cls)

    def get_label(self):
        if getattr(self, "label", ""):
            return self.label
        else:
            return to_title_case(self.name)


@as_discriminated_union
class Component(Labelled):
    """These make up a Device"""


@dataclass
class SignalR(Component):
    """Scalar value backed by a single PV"""

    pv: Annotated[str, desc("PV to be used for get and monitor")]
    widget: Annotated[
        Optional[ReadWidget],
        desc("Widget to use for display, None means don't display"),
    ] = None


@dataclass
class SignalW(Component):
    """Write only value backed by a single PV"""

    pv: Annotated[str, desc("PV to be used for put")]
    widget: Annotated[
        Optional[WriteWidget],
        desc("Widget to use for control, None means don't display"),
    ] = None


@dataclass
class SignalRW(Component):
    """Read/write value backed by one or two PVs"""

    pv: Annotated[str, desc("PV to be used for put")]
    widget: Annotated[
        Optional[WriteWidget],
        desc("Widget to use for control, None means don't display"),
    ] = None
    # This was Optional[str] but produced JSON schema that YAML editor didn't understand
    read_pv: Annotated[str, desc("PV to be used for read, empty means use pv")] = ""
    read_widget: Annotated[
        Optional[ReadWidget], desc("Widget to use for display, None means use widget")
    ] = None


SignalTypes = (SignalR, SignalW, SignalRW)
ReadSignalType = Union[SignalR, SignalRW]
WriteSignalType = Union[SignalW, SignalRW]


@dataclass
class SignalX(Component):
    """Executable that puts a fixed value to a PV."""

    pv: Annotated[str, desc("PV to be used for call")]
    value: Annotated[Any, desc("Value to write. None means zero")] = None


@dataclass
class DeviceRef(Component):
    """Reference to another Device."""

    pv: Annotated[str, desc("Child device PVI PV")]
    ui: Annotated[str, desc("UI file to open for referenced Device")]
    macros: Annotated[
        Dict[str, str],
        desc("Macro-value pairs for UI file"),
    ] = field(default_factory=dict)


class SignalRef(Component):
    """Reference to another Signal with the same name in this Device."""


T = TypeVar("T")
S = TypeVar("S")


@add_type_field
@dataclass
class Group(Generic[T], Labelled):
    """Group of child components in a Layout"""

    layout: Annotated[Layout, desc("How to layout children on screen")]
    children: Annotated[Tree[T], desc("Child Components")]


Tree = Sequence[Union[T, Group[T]]]


def on_each_node(tree: Tree[T], func: Callable[[T], Iterator[S]]) -> Tree[S]:
    """Visit each node of the tree of type typ, calling func on each leaf"""
    out: List[Union[S, Group[S]]] = []
    for t in tree:
        if isinstance(t, Group):
            group: Group[S] = Group(
                name=t.name, layout=t.layout, children=on_each_node(t.children, func)
            )
            out.append(group)
        else:
            out += list(func(t))
    return out


def walk(tree: Tree[T]) -> Iterator[Union[T, Group[T]]]:
    """Depth first traversal of tree"""
    for t in tree:
        yield t
        if isinstance(t, Group):
            yield from walk(t.children)


def signal_access_mode(signal: SignalR | SignalW | SignalRW):
    match signal:
        case SignalR():
            return "r"
        case SignalW():
            return "w"
        case SignalRW():
            return "rw"


@dataclass
class Device:
    """Collection of Components"""

    label: Annotated[str, desc("Label for screen")]
    parent: Optional[
        Annotated[
            str,
            desc("The parent device (basename of yaml file)"),
        ]
    ] = None
    children: Annotated[Tree[Component], desc("Child Components")] = field(
        default_factory=list
    )

    def serialize(self) -> Mapping[str, Any]:
        """Serialize the Device to a dictionary."""
        return serialize(self, exclude_none=True, exclude_defaults=True)

    @classmethod
    def deserialize(cls, serialized: Path) -> Device:
        """Deserialize the Device from a YAML file"""
        if not serialized.is_file():
            raise FileNotFoundError(serialized)

        return deserialize(cls, YAML(typ="safe").load(serialized))

    def deserialize_parents(self, yaml_paths: List[Path]):
        """Deserialize yaml of parents and extract parameters"""
        if self.parent is None or self.parent == "asynPortDriver":
            return

        parent_parameters = find_components(self.parent, yaml_paths)
        for node in parent_parameters:
            if isinstance(node, Group):
                for param_group in self.children:
                    if not isinstance(param_group, Group):
                        continue
                    elif param_group.name == node.name:
                        param_group.children = list(node.children) + list(
                            param_group.children
                        )
                        break  # Groups merged - skip to next parent group

                else:  # No break - Did not find the Group
                    # Inherit as a new Group
                    self.children = list(self.children) + [node]
                    continue  # Skip to next parent group

            else:
                # Node is an individual AsynParameter - just append it
                self.children = list(self.children) + [node]

    def generate_param_tree(self) -> str:
        param_tree = ", ".join(json.dumps(serialize(group)) for group in self.children)
        # Encode again to quote the string as a value and escape double quotes within
        return json.dumps('{"parameters":[' + param_tree + "]}")


def find_components(yaml_name: str, yaml_paths: List[Path]) -> Tree[Component]:
    if yaml_name == "asynPortDriver":
        return []  # asynPortDriver is the most base class and has no parameters

    # Look in this module first
    device_name = f"{yaml_name}.pvi.device.yaml"
    device_yaml = find_pvi_yaml(device_name, yaml_paths)

    if device_yaml is None:
        raise IOError(f"Cannot find {device_name} in {yaml_paths}")

    device = Device.deserialize(device_yaml)

    parent_components = (
        list(find_components(device.parent, yaml_paths))
        if device.parent is not None
        else []
    )

    return list(device.children) + parent_components



import re
from dataclasses import dataclass
from typing import Optional, Type

from pvi.device import Component, SignalR, SignalRW, SignalW

from ._parameters import (
    AsynParameter,
    InRecordTypes,
    OutRecordTypes,
    Parameter,
    Record,
    get_waveform_parameter,
)


class RecordError(Exception):
    pass


@dataclass
class AsynRecord(Record):
    def __post_init__(self):
        # We don't care about records without INP or OUT or with both (error)
        if all(k in self.fields.keys() for k in ("INP", "OUT")) or not any(
            k in self.fields.keys() for k in ("INP", "OUT")
        ):
            raise RecordError("Record has no input or output field or both")

        asyn_field = self.fields.get("INP", self.fields.get("OUT"))
        if "@asyn(" not in asyn_field:
            # Is it Asyn, because if it's not we don't care about it
            raise RecordError("Record has no @asyn field")

        # If there is no DESC field we must create one
        if "DESC" not in self.fields.keys():
            self.fields["DESC"] = self.name

    def get_parameter_name(self) -> Optional[str]:
        # e.g. from: field(INP,  "@asyn($(PORT),$(ADDR=0),$(TIMEOUT=1))FILE_PATH")
        # extract: FILE_PATH
        parameter_name_extractor = r"(?:@asyn\()(?:(?:\$\([^\)]*\)[,]*)*)(?:\))(\S+)"
        parameter_name = None
        for k, v in self.fields.items():
            if k == "INP" or k == "OUT":
                if "@asyn(" in v:
                    match = re.search(parameter_name_extractor, v)
                    if match:
                        parameter_name = match.group(1)
        return parameter_name

    def asyn_component_type(self) -> Type[AsynParameter]:
        # For waveform records the data type is defined by DTYP
        if self.type == "waveform":
            return get_waveform_parameter(self.fields["DTYP"])

        try:
            if "INP" in self.fields.keys():
                record_types = InRecordTypes
                return record_types[self.type]
            elif "OUT" in self.fields.keys():
                record_types = OutRecordTypes
                return record_types[self.type]
        except KeyError as e:
            raise KeyError(
                f"{self.name} asyn type {self.type}({self.fields}) not found in"
                f"{list(record_types)}"
            ) from e

        raise ValueError(
            f"Could not determine asyn type for {self.name} fields {self.fields}"
        )


@dataclass
class SettingPair(Parameter):
    read_record: AsynRecord
    write_record: AsynRecord

    def generate_component(self) -> SignalRW:
        asyn_cls = self.write_record.asyn_component_type()
        component = asyn_cls(name=self.write_record.name)

        return SignalRW(
            name=component.name,
            pv=component.get_write_record(),
            widget=component.write_widget,
            read_pv=component.get_read_record(),
            read_widget=component.read_widget,
        )


@dataclass
class Readback(Parameter):
    read_record: AsynRecord

    def generate_component(self) -> SignalR:
        asyn_cls = self.read_record.asyn_component_type()

        if self.read_record.name.endswith("_RBV"):
            name = self.read_record.name[: -len("_RBV")]
        else:
            name = self.read_record.name

        component = asyn_cls(name=name, read_record=self.read_record.name)

        return SignalR(
            name=component.name,
            pv=component.get_read_record(),
            widget=component.read_widget,
        )


@dataclass
class Action(Parameter):
    write_record: AsynRecord

    def generate_component(self) -> Component:
        asyn_cls = self.write_record.asyn_component_type()
        component = asyn_cls(self.write_record.name)

        return SignalW(
            name=component.name,
            pv=component.get_write_record(),
            widget=component.write_widget,
        )



import re
from pathlib import Path
from typing import Type, TypeVar

from apischema import deserialize, serialize
from ruamel.yaml import YAML

from ._schema_utils import has_type

T = TypeVar("T")


def add_line_before_type(s: str) -> str:
    return re.sub(r"(\s*- type:)", "\n\\g<1>", s)


def serialize_yaml(obj, path: Path):
    serialized = serialize(obj, exclude_none=True, exclude_defaults=True)
    # TODO: add modeline
    YAML().dump(serialized, path, transform=add_line_before_type)


def deserialize_yaml(cls: Type[T], path: Path) -> T:
    # Walk up to find the deserialization root
    while cls not in has_type and len(cls.__mro__) > 1:
        cls = cls.__mro__[1]
    suffix = f".pvi.{cls.__name__.lower()}.yaml"
    assert path.name.endswith(suffix), f"Expected '{path.name}' to end with '{suffix}'"
    # Need to use the safe loader otherwise we get:
    #    TypeError: Invalid JSON type <class 'ruamel.yaml.scalarfloat.ScalarFloat'>
    d = YAML(typ="safe").load(path)
    return deserialize(cls, d)



from pathlib import Path

import pytest

from pvi._format.base import Formatter, IndexEntry
from pvi._format.dls import DLSFormatter
from pvi._format.template import format_template
from pvi._yaml_utils import deserialize_yaml
from pvi.device import (
    LED,
    ButtonPanel,
    ComboBox,
    Device,
    DeviceRef,
    Grid,
    Group,
    SignalR,
    SignalRW,
    SignalW,
    SubScreen,
    TableRead,
    TableWrite,
    TextFormat,
    TextRead,
    TextWrite,
)

HERE = Path(__file__).parent


@pytest.mark.parametrize(
    "filename,formatter",
    [
        ("text_format.adl", "aps.adl.pvi.formatter.yaml"),
        ("text_format.edl", "dls.edl.pvi.formatter.yaml"),
        ("text_format.bob", "dls.bob.pvi.formatter.yaml"),
    ],
)
def test_text_format(tmp_path, helper, filename, formatter):
    formatter_yaml = HERE / "format" / "input" / formatter
    formatter = deserialize_yaml(Formatter, formatter_yaml)

    components = []
    for format in TextFormat:
        components.append(
            SignalRW(
                format.name.title(),
                pv=format.name.title(),
                widget=TextWrite(format=format),
                read_pv=f"{format.name.title()}_RBV",
                read_widget=TextRead(format=format),
            )
        )

    device = Device("Text Device", children=components)

    expected_ui = HERE / "format" / "output" / filename
    output_ui = tmp_path / filename
    formatter.format(device, "$(P)", output_ui)

    helper.assert_output_matches(expected_ui, output_ui)


def test_button(tmp_path, helper):
    formatter_yaml = HERE / "format" / "input" / "dls.bob.pvi.formatter.yaml"
    formatter = deserialize_yaml(Formatter, formatter_yaml)

    acquire_time = SignalRW(
        "AcquireTime",
        pv="AcquireTime",
        widget=TextWrite(),
        read_pv="AcquireTime_RBV",
        read_widget=TextRead(),
    )
    acquire = SignalW(
        "Acquire",
        pv="Acquire",
        widget=ButtonPanel(actions=dict(Start=1, Stop=0)),
    )
    acquire_w_rbv = SignalRW(
        "AcquireWithRBV",
        pv="Acquire",
        widget=ButtonPanel(actions=dict(Start=1, Stop=0)),
        read_pv="Acquire_RBV",
        read_widget=LED(),
    )
    device = Device("Simple Device", children=[acquire_time, acquire, acquire_w_rbv])

    expected_bob = HERE / "format" / "output" / "button.bob"
    output_bob = tmp_path / "button.bob"
    formatter.format(device, "$(P)", output_bob)

    helper.assert_output_matches(expected_bob, output_bob)


def test_pva_table(tmp_path, helper):
    formatter_yaml = HERE / "format" / "input" / "dls.bob.pvi.formatter.yaml"
    formatter = deserialize_yaml(Formatter, formatter_yaml)

    table = SignalR(
        "PVATable",
        "PVATable",
        TableWrite([TextWrite(), LED(), ComboBox(["1", "A", "True"])]),
    )
    device = Device("TableDevice", children=[table])

    expected_bob = HERE / "format" / "output" / "pva_table.bob"
    output_bob = tmp_path / "pva_table.bob"
    formatter.format(device, "$(P)$(R)", output_bob)

    helper.assert_output_matches(expected_bob, output_bob)


def test_pva_table_panda(tmp_path, helper):
    formatter_yaml = HERE / "format" / "input" / "dls.bob.pvi.formatter.yaml"
    formatter = deserialize_yaml(Formatter, formatter_yaml)

    table = SignalR(
        "PandA ",
        "PANDAQSRV:SEQ1:TABLE",
        TableRead([TextRead()] * 4 + [LED()] * 6 + [TextRead()] + [LED()] * 6),
    )
    device = Device("TableDevice", children=[table])

    expected_bob = HERE / "format" / "output" / "pva_table_panda.bob"
    output_bob = tmp_path / "pva_table_panda.bob"
    formatter.format(device, "", output_bob)

    helper.assert_output_matches(expected_bob, output_bob)


def test_combo_box(tmp_path, helper):
    formatter_yaml = HERE / "format" / "input" / "dls.bob.pvi.formatter.yaml"
    formatter = deserialize_yaml(Formatter, formatter_yaml)

    combo_box = SignalR("Enum", "Enum", ComboBox(["A", "B", "C"]))
    device = Device("Device", children=[combo_box])

    expected_bob = HERE / "format" / "output" / "combo_box.bob"
    output_bob = tmp_path / "combo_box.bob"
    formatter.format(device, "$(P)$(R)", output_bob)

    helper.assert_output_matches(expected_bob, output_bob)


def test_device_ref(tmp_path, helper):
    formatter_yaml = HERE / "format" / "input" / "dls.bob.pvi.formatter.yaml"
    formatter = deserialize_yaml(Formatter, formatter_yaml)

    # Make a button to open an existing screen - use screen from combo box test
    device_ref = DeviceRef(
        "ComboBox", pv="COMBOBOX", ui="combo_box", macros=dict(P="EIGER", R=":CAM:")
    )
    device = Device("Device", children=[device_ref])

    expected_bob = HERE / "format" / "output" / "device_ref.bob"
    output_bob = tmp_path / "device_ref.bob"
    formatter.format(device, "$(P)$(R)", output_bob)

    helper.assert_output_matches(expected_bob, output_bob)


def test_group_sub_screen(tmp_path, helper):
    formatter_yaml = HERE / "format" / "input" / "dls.bob.pvi.formatter.yaml"
    formatter = deserialize_yaml(Formatter, formatter_yaml)

    # This tests every valid combination of nesting Group and SubScreen
    signals = [
        SignalR("A", "A", widget=TextRead()),
        Group(
            "Group 1",
            layout=SubScreen(),
            children=[
                SignalR("Group 1 B", "GROUP1:B", widget=TextRead()),
                Group(
                    "Group 2",
                    layout=Grid(),
                    children=[
                        SignalR("Group 2 C", "GROUP1:GROUP2:C", widget=TextRead()),
                        SignalR("Group 2 D", "GROUP1:GROUP2:D", widget=TextRead()),
                    ],
                ),
            ],
        ),
        Group(
            "Group 3",
            layout=Grid(),
            children=[
                SignalR("Group 3 E", "GROUP3:E", widget=TextRead()),
                Group(
                    "Group 4",
                    layout=SubScreen(),
                    children=[
                        SignalR("Group 4 F", "GROUP3:GROUP4:F", widget=TextRead()),
                        Group(
                            "Group 5",
                            layout=Grid(),
                            children=[
                                SignalR(
                                    "Group 5 G",
                                    "GROUP3:GROUP4:GROUP5:G",
                                    widget=TextRead(),
                                ),
                                SignalR(
                                    "Group 5 H",
                                    "GROUP3:GROUP4:GROUP5:H",
                                    widget=TextRead(),
                                ),
                            ],
                        ),
                    ],
                ),
            ],
        ),
    ]
    device = Device("Device", children=signals)

    expected_bob = HERE / "format" / "output" / "sub_screen.bob"
    output_bob = tmp_path / "sub_screen.bob"
    formatter.format(device, "$(P)$(R)", output_bob)

    helper.assert_output_matches(expected_bob, output_bob)


def test_index(tmp_path, helper):
    expected_bob = HERE / "format" / "output" / "index.bob"
    output_bob = tmp_path / "index.bob"

    DLSFormatter().format_index(
        "Index",
        [
            IndexEntry("Button", "button.bob", {"P": "BUTTON"}),
            IndexEntry("ComboBox", "combo_box.bob", {"P": "COMBOBOX"}),
            IndexEntry("Table", "pva_table.bob", {"P": "TABLE"}),
        ],
        output_bob,
    )

    helper.assert_output_matches(expected_bob, output_bob)


def test_pvi_template(tmp_path, helper):
    read = SignalR("Read", "Read")
    write = SignalW("Write", "Write")
    read_write = SignalRW("ReadWrite", "ReadWrite")
    device = Device("Template Device", children=[read, write, read_write])

    expected_bob = HERE / "format" / "output" / "pvi.template"
    output_template = tmp_path / "pvi.template"
    format_template(device, "$(P)$(R)", output_template)

    helper.assert_output_matches(expected_bob, output_template)



import json
from pathlib import Path

import pytest
from ruamel.yaml import YAML

from pvi.device import (
    LED,
    CheckBox,
    ComboBox,
    Device,
    Grid,
    Group,
    SignalR,
    SignalRW,
    TableWrite,
    TextRead,
    TextWrite,
)


@pytest.fixture
def device():
    components = [
        Group(
            "Parameters",
            Grid(),
            [
                SignalRW("WidthUnits", "WIDTH:UNITS", ComboBox()),
                SignalRW("Width", "WIDTH", TextWrite(), "WIDTH_RBV", TextRead()),
            ],
        ),
        SignalRW("Table", "TABLE", TableWrite([CheckBox(), ComboBox(), TextWrite()])),
        SignalR("OutA", "OUTA", LED()),
    ]
    return Device("label", "parent", components)


DEVICE_YAML = Path(__file__).parent / "test_device_serialization.pvi.yaml"


def test_serialize(device: Device):
    s = device.serialize()
    assert json.dumps(s, indent=2) == json.dumps(
        YAML(typ="safe").load(DEVICE_YAML), indent=2
    )


def test_deserialize(device: Device):
    d = Device.deserialize(DEVICE_YAML)
    assert d == device



from dataclasses import field, make_dataclass
from functools import lru_cache
from typing import (
    Any,
    Callable,
    List,
    Literal,
    Mapping,
    Optional,
    Pattern,
    Set,
    TypeVar,
    Union,
)

from apischema import deserializer, order, schema, serialized, type_name
from apischema.conversions import Conversion
from apischema.conversions.converters import serializer
from apischema.json_schema import JsonSchemaVersion, deserialization_schema
from apischema.utils import identity

Cls = TypeVar("Cls", bound=type)


# Permanently cache so we don't include deserialization subclasses defined below
@lru_cache(maxsize=None)
def rec_subclasses(cls: Cls) -> List[Cls]:
    """Recursive implementation of type.__subclasses__"""
    subclasses = []
    for sub_cls in cls.__subclasses__():
        subclasses += [sub_cls] + rec_subclasses(sub_cls)
    return subclasses


has_type: Set[type] = set()


def _make_converters(cls: Cls, classes: Callable[[Cls], List[Cls]]) -> Cls:
    params = tuple(getattr(cls, "__parameters__", ()))

    def with_params(sub: Cls) -> Any:
        return sub[params] if params else sub  # type: ignore

    def deserialization() -> Conversion:
        def typed_subs():
            for sub in classes(cls):
                # Make typed_sub derived from sub with additional type entry
                type_field = (
                    "type",
                    Literal[sub.__name__],  # type: ignore
                    field(default=sub.__name__, metadata=order(-1)),
                )
                yield make_dataclass(
                    sub.__name__, [type_field], bases=(with_params(sub),)
                )

        def convert(sub_inst):
            """Make instance of sub from instance of typed_sub"""
            d = {k: v for k, v in sub_inst.__dict__.items() if k != "type"}
            return type(sub_inst).__bases__[0](**d)

        return Conversion(
            convert,
            source=Union[tuple(with_params(s) for s in typed_subs())],  # type: ignore
            target=with_params(cls),
        )

    def serialization() -> Conversion:
        @serialized("type", owner=cls, order=order(-1))
        def add_type(obj) -> str:
            """Add type entry to serialized dictionary"""
            return type(obj).__name__

        return Conversion(
            identity,
            source=cls,
            target=Union[tuple(classes(cls))],  # type: ignore
            inherited=False,
        )

    serializer(lazy=serialization, source=cls)
    deserializer(lazy=deserialization, target=cls)
    type_name(lambda tp, arg: f"{arg.__name__}{tp.__name__}")(cls)
    has_type.add(cls)
    return cls


def add_type_field(cls: Cls) -> Cls:
    return _make_converters(cls, lambda cls: [cls])


def as_discriminated_union(cls: Cls) -> Cls:
    return _make_converters(cls, rec_subclasses)


def desc(description: str, *, pattern: Optional[Union[str, Pattern]] = None):
    return schema(description=description, pattern=pattern)


def make_json_schema(cls: Cls) -> Mapping[str, Any]:
    return deserialization_schema(cls, all_refs=True, version=JsonSchemaVersion.DRAFT_7)



import json
import os
from pathlib import Path
from typing import List, Optional, Type

import typer

from pvi import __version__
from pvi._convert._template_convert import TemplateConverter
from pvi._convert.utils import extract_device_and_parent_class
from pvi._format import Formatter
from pvi._format.template import format_template
from pvi._pv_group import group_parameters
from pvi._schema_utils import make_json_schema
from pvi._yaml_utils import deserialize_yaml, serialize_yaml
from pvi.device import Device

app = typer.Typer()
convert_app = typer.Typer()
app.add_typer(convert_app, name="convert", help="Convert a module to use PVI")


def version_callback(value: bool):
    if value:
        typer.echo(__version__)
        raise typer.Exit()


@app.callback()
def main(
    version: Optional[bool] = typer.Option(
        None,
        "--version",
        callback=version_callback,
        is_eager=True,
        help="Print the version and exit",
    )
):
    """PVI builder interface"""


@app.command()
def schema(output: Path = typer.Argument(..., help="filename to write the schema to")):
    """Write the JSON schema for the pvi interface"""
    cls: Type
    assert output.name.endswith(
        ".schema.json"
    ), f"Expected '{output.name}' to end with '.schema.json'"
    if output.name == "pvi.device.schema.json":
        cls = Device
    elif output.name == "pvi.formatter.schema.json":
        cls = Formatter
    else:
        typer.echo(f"Don't know how to create {output.name}")
        raise typer.Exit(code=1)
    schema = make_json_schema(cls)
    output.write_text(json.dumps(schema, indent=2))


@app.command()
def format(
    output_path: Path = typer.Argument(
        ..., help="Directory to write output file(s) to"
    ),
    device_path: Path = typer.Argument(..., help="Path to the .pvi.device.yaml file"),
    formatter_path: Path = typer.Argument(
        ..., help="Path to the .pvi.formatter.yaml file"
    ),
    yaml_paths: List[Path] = typer.Option(
        [], "--yaml-path", help="Paths to directories with .pvi.device.yaml files"
    ),
):
    """Create screen product from device and formatter YAML"""
    device = Device.deserialize(device_path)
    device.deserialize_parents(yaml_paths)

    formatter_inst: Formatter = deserialize_yaml(Formatter, formatter_path)
    formatter_inst.format(device, "$(P)$(R)", output_path)


@app.command()
def generate_template(
    device_path: Path = typer.Argument(..., help="Path to the .pvi.device.yaml file"),
    prefix: str = typer.Argument(..., help="PV Prefix in generated template"),
    output_path: Path = typer.Argument(..., help="Output file to generate"),
):
    """Create template with info tags for device signals"""
    device = Device.deserialize(device_path)
    format_template(device, prefix, output_path)


@convert_app.command()
def device(
    output: Path = typer.Argument(..., help="Directory to write output file to"),
    h: Path = typer.Argument(..., help="Path to the .h file to convert"),
    templates: List[Path] = typer.Option(
        [], "--template", help="Paths to .template files to convert"
    ),
):
    """Convert template to device YAML"""
    if not output.exists():
        os.mkdir(output)

    name, parent = extract_device_and_parent_class(h.read_text())
    component_group = TemplateConverter(templates).convert()
    device = Device(name, parent, component_group)

    serialize_yaml(device, output / f"{name}.pvi.device.yaml")


@app.command()
def regroup(
    device_path: Path = typer.Argument(
        ..., help="Path to the device.yaml file to regroup"
    ),
    ui_paths: List[Path] = typer.Argument(
        ..., help="Paths to the ui files to regroup the PVs by"
    ),
):
    """Regroup a device.yaml file based on ui files that the PVs appear in"""
    device = Device.deserialize(device_path)
    device.children = group_parameters(device, ui_paths)

    serialize_yaml(device, device_path)


# test with: pipenv run python -m pvi
if __name__ == "__main__":
    app()



import re
from pathlib import Path
from typing import Dict, List, Tuple

from pvi.device import Component, Device, Grid, Group, walk


def find_pvs(pvs: List[str], file_path: Path) -> Tuple[List[str], List[str]]:
    """Search for the PVs in the file and return lists of found and not found pvs"""

    with open(file_path, "r") as f:
        file_content = f.read()

    pv_coordinates: Dict[int, List[str]] = {}
    remaining_pvs = [pv for pv in pvs]
    for pv in pvs:
        if pv not in file_content:
            continue

        # This regex searches for the first y coordinate before the given PV
        # https://regex101.com/r/41yX56/1
        widget_regex = re.compile(
            "".join(
                (
                    # The y-coordinate in a match group named `y`
                    r"y=(?P<y>\d+)",
                    # Negative lookahead asserting no further y-coordinate before PV
                    r"(?!.*y=.*" + pv + r")",
                    # Any characters and then PV
                    r".*" + pv,
                )
            ),
            re.MULTILINE | re.DOTALL,
        )
        match = re.search(widget_regex, file_content)

        assert match, f"{pv} found in {file_path.name} but did not match a y coordinate"

        y = int(match["y"])
        if y in pv_coordinates:
            pv_coordinates[y].append(pv)
        else:
            pv_coordinates[y] = [pv]

        remaining_pvs.remove(pv)

    grouped_pvs = []
    for coord in sorted(pv_coordinates.keys()):
        grouped_pvs.extend(pv_coordinates[coord])

    return grouped_pvs, remaining_pvs


def sanitize_name(name: str) -> str:
    name = name.replace(" ", "")
    name = name.replace("-", "")
    name = name.replace("_", "")
    return name


def group_parameters(device: Device, ui_paths: List[Path]) -> List[Group[Component]]:
    initial_parameters: List[Component] = [
        param for param in walk(device.children) if isinstance(param, Component)
    ]

    # Group PVs that appear in the given ui files
    pvs = [node.name for node in walk(device.children) if isinstance(node, Component)]

    group_pv_map: Dict[str, List[str]] = {}
    for ui in ui_paths:
        ui_pvs, pvs = find_pvs(pvs, ui)
        if ui_pvs:
            group_pv_map[ui.stem] = ui_pvs

    if pvs:
        print(f'Did not find group for {"|".join(pvs)}')

    # Create groups for parameters we found in the files
    ui_groups: List[Group[Component]] = [
        Group(
            sanitize_name(group_name),
            Grid(labelled=True),
            [  # Note: Need to preserve order in group_pvs here
                param
                for pv in group_pvs
                for param in initial_parameters
                if param.name == pv
            ],
        )
        for group_name, group_pvs in group_pv_map.items()
    ]

    # Separate any parameters we failed to find a group for
    grouped_parameters = [param for group in ui_groups for param in group.children]
    ungrouped_parameters = [
        param for param in initial_parameters if param not in grouped_parameters
    ]

    # Replace with grouped parameters and any ungrouped parameters on the end
    if ungrouped_parameters:
        ui_groups.append(
            Group(
                sanitize_name(device.label + "Misc"),
                Grid(labelled=True),
                ungrouped_parameters,
            )
        )

    return ui_groups



import subprocess
import sys
from pathlib import Path

import pytest

from pvi import __version__
from pvi.__main__ import app

HERE = Path(__file__).parent


def test_cli_version():
    cmd = [sys.executable, "-m", "pvi", "--version"]
    assert subprocess.check_output(cmd).decode().strip() == __version__


@pytest.mark.parametrize(
    "filename",
    [
        "pvi.device.schema.json",
        "pvi.formatter.schema.json",
    ],
)
def test_schemas(tmp_path, helper, filename):
    expected_path = HERE.parent / "schemas" / filename
    helper.assert_cli_output_matches(app, expected_path, "schema", tmp_path / filename)


@pytest.mark.parametrize(
    "filename,formatter",
    [
        ("mixedWidgets.adl", "aps.adl.pvi.formatter.yaml"),
        ("mixedWidgets.edl", "dls.edl.pvi.formatter.yaml"),
        ("mixedWidgets.bob", "dls.bob.pvi.formatter.yaml"),
    ],
)
def test_format(tmp_path, helper, filename, formatter):
    expected_path = HERE / "format" / "output" / filename
    input_path = HERE / "format" / "input"
    formatter_path = input_path / formatter
    helper.assert_cli_output_matches(
        app,
        expected_path,
        "format --yaml-path " + str(input_path),
        tmp_path / filename,
        HERE / "format" / "input" / "mixedWidgets.pvi.device.yaml",
        formatter_path,
    )


def test_convert(tmp_path, helper):
    expected_path = HERE / "convert" / "output"
    input_path = HERE / "convert" / "input"
    helper.assert_cli_output_matches(
        app,
        expected_path,
        "convert device",
        tmp_path,
        input_path / "simDetector.h",
        "--template",
        input_path / "simDetector.template",
    )


@pytest.mark.parametrize(
    "input_yaml,formatter,output",
    [
        (
            "static_table.pvi.device.yaml",
            "dls.edl.pvi.formatter.yaml",
            "static_table.edl",
        ),
        (
            "static_table.pvi.device.yaml",
            "dls.bob.pvi.formatter.yaml",
            "static_table.bob",
        ),
        (
            "static_table.pvi.device.yaml",
            "aps.pvi.formatter.yaml",
            "static_table.adl",
        ),
    ],
)
def test_static_table(tmp_path, helper, input_yaml, formatter, output):
    expected_path = HERE / "format" / "output" / output
    input_path = HERE / "format" / "input"
    formatter_path = HERE / "../formatters/" / formatter
    helper.assert_cli_output_matches(
        app,
        expected_path,
        "format --yaml-path " + str(input_path),
        tmp_path / output,
        input_path / input_yaml,
        formatter_path,
    )



from dataclasses import dataclass
from pathlib import Path
from typing import Annotated, List

from lxml import etree

from pvi._format.bob import BobTemplate
from pvi._format.edl import EdlTemplate
from pvi._format.screen import (
    ScreenFormatterFactory,
    ScreenLayout,
    WidgetFormatterFactory,
)
from pvi._format.widget import (
    ActionWidgetFormatter,
    GroupFormatter,
    LabelWidgetFormatter,
    PVWidgetFormatter,
    SubScreenWidgetFormatter,
    WidgetFormatter,
)
from pvi._schema_utils import desc
from pvi.device import Device

from .base import Formatter
from .utils import Bounds, GroupType, with_title


@dataclass
class DLSFormatter(Formatter):
    spacing: Annotated[int, desc("Spacing between widgets")] = 5
    title_height: Annotated[int, desc("Height of screen title bar")] = 25
    max_height: Annotated[int, desc("Max height of the screen")] = 900
    label_width: Annotated[int, desc("Width of the widget description labels")] = 115
    widget_width: Annotated[int, desc("Width of the widgets")] = 120
    widget_height: Annotated[int, desc("Height of the widgets")] = 20

    def format(self, device: Device, prefix: str, path: Path):
        if path.suffix == ".edl":
            f = self.format_edl
        elif path.suffix == ".bob":
            f = self.format_bob
        else:
            raise ValueError("Can only write .edl or .bob files")
        f(device, prefix, path)

    def format_edl(self, device: Device, prefix: str, path: Path):
        template = EdlTemplate((Path(__file__).parent / "dls.edl").read_text())
        screen_layout = ScreenLayout(
            spacing=self.spacing,
            title_height=self.title_height,
            max_height=self.max_height,
            group_label_height=10,
            label_width=self.label_width,
            widget_width=self.widget_width,
            widget_height=self.widget_height,
            group_widget_indent=5,
            group_width_offset=0,
        )
        widget_formatter_factory = WidgetFormatterFactory(
            header_formatter_cls=LabelWidgetFormatter.from_template(
                template,
                search='"Heading"',
                property_map=dict(value="text"),
            ),
            label_formatter_cls=LabelWidgetFormatter.from_template(
                template,
                search='"Label"',
                property_map=dict(value="text"),
            ),
            led_formatter_cls=PVWidgetFormatter.from_template(
                template,
                search='"LED"',
                sized=Bounds.square,
                property_map=dict(controlPv="pv"),
            ),
            progress_bar_formatter_cls=PVWidgetFormatter.from_template(
                template,
                search='"ProgressBar"',
                property_map=dict(indicatorPv="pv"),
            ),
            text_read_formatter_cls=PVWidgetFormatter.from_template(
                template,
                search='"TextRead"',
                property_map=dict(controlPv="pv"),
            ),
            check_box_formatter_cls=PVWidgetFormatter.from_template(
                template,
                search='"ComboBox"',
                property_map=dict(controlPv="pv"),
            ),
            combo_box_formatter_cls=PVWidgetFormatter.from_template(
                template,
                search='"ComboBox"',
                property_map=dict(controlPv="pv"),
            ),
            text_write_formatter_cls=PVWidgetFormatter.from_template(
                template,
                search='"TextWrite"',
                property_map=dict(controlPv="pv"),
            ),
            # Cannot handle dynamic tables so insert a label with the PV name
            table_formatter_cls=PVWidgetFormatter.from_template(
                template,
                search='"Label"',
                property_map=dict(value="pv"),
            ),
            action_formatter_cls=ActionWidgetFormatter.from_template(
                template,
                search='"SignalX"',
                property_map=dict(onLabel="label", offLabel="label", controlPv="pv"),
            ),
            sub_screen_formatter_cls=SubScreenWidgetFormatter.from_template(
                template,
                search='"SubScreenFile"',
                property_map=dict(displayFileName="file_name"),
            ),
        )
        screen_title_cls = LabelWidgetFormatter.from_template(
            template,
            search='"Title"',
            property_map=dict(value="text"),
        )
        group_title_cls = LabelWidgetFormatter.from_template(
            template,
            search='"  Group  "',
            property_map=dict(value="text"),
        )
        group_box_cls = WidgetFormatter.from_template(
            template, search="fillColor index 5"
        )

        def create_group_box_formatter(
            bounds: Bounds, title: str
        ) -> List[WidgetFormatter[str]]:
            x, y, w, h = bounds.x, bounds.y, bounds.w, bounds.h
            return [
                group_box_cls(
                    Bounds(
                        x,
                        y + screen_layout.spacing,
                        w,
                        h - screen_layout.spacing,
                    )
                ),
                group_title_cls(
                    Bounds(x, y, w, screen_layout.group_label_height),
                    f"  {title}  ",
                ),
            ]

        def create_screen_title_formatter(
            bounds: Bounds, title: str
        ) -> List[WidgetFormatter[str]]:
            return [
                screen_title_cls(
                    Bounds(0, 0, bounds.w, screen_layout.title_height), title
                )
            ]

        formatter_factory = ScreenFormatterFactory(
            screen_formatter_cls=GroupFormatter.from_template(
                template,
                search=GroupType.SCREEN,
                sized=with_title(screen_layout.spacing, screen_layout.title_height),
                widget_formatter_hook=create_screen_title_formatter,
            ),
            group_formatter_cls=GroupFormatter.from_template(
                template,
                search=GroupType.GROUP,
                sized=with_title(
                    screen_layout.spacing, screen_layout.group_label_height
                ),
                widget_formatter_hook=create_group_box_formatter,
            ),
            widget_formatter_factory=widget_formatter_factory,
            prefix=prefix,
            layout=screen_layout,
            base_file_name=path.stem,
        )
        title = f"{device.label} - {prefix}"

        screen_formatter, sub_screens = formatter_factory.create_screen_formatter(
            device.children, title
        )

        path.write_text("".join(screen_formatter.format()))
        for sub_screen_name, sub_screen_formatter in sub_screens:
            sub_screen_path = Path(path.parent / f"{sub_screen_name}{path.suffix}")
            sub_screen_path.write_text("".join(sub_screen_formatter.format()))

    def format_bob(self, device: Device, prefix: str, path: Path):
        template = BobTemplate(str(Path(__file__).parent / "dls.bob"))
        # LP DOCS REF: Define the layout properties
        screen_layout = ScreenLayout(
            spacing=self.spacing,
            title_height=self.title_height,
            max_height=self.max_height,
            group_label_height=26,
            label_width=self.label_width,
            widget_width=self.widget_width,
            widget_height=self.widget_height,
            group_widget_indent=18,
            group_width_offset=26,
        )
        # SW DOCS REF: Extract widget types from template file
        widget_formatter_factory = WidgetFormatterFactory(
            header_formatter_cls=LabelWidgetFormatter.from_template(
                template,
                search="Heading",
                property_map=dict(text="text"),
            ),
            label_formatter_cls=LabelWidgetFormatter.from_template(
                template,
                search="Label",
                property_map=dict(text="text"),
            ),
            led_formatter_cls=PVWidgetFormatter.from_template(
                template,
                search="LED",
                sized=Bounds.square,
                property_map=dict(pv_name="pv"),
            ),
            progress_bar_formatter_cls=PVWidgetFormatter.from_template(
                template,
                search="ProgressBar",
                property_map=dict(pv_name="pv"),
            ),
            text_read_formatter_cls=PVWidgetFormatter.from_template(
                template,
                search="TextUpdate",
                property_map=dict(pv_name="pv"),
            ),
            check_box_formatter_cls=PVWidgetFormatter.from_template(
                template,
                search="ChoiceButton",
                property_map=dict(pv_name="pv"),
            ),
            combo_box_formatter_cls=PVWidgetFormatter.from_template(
                template,
                search="ComboBox",
                property_map=dict(pv_name="pv"),
            ),
            text_write_formatter_cls=PVWidgetFormatter.from_template(
                template,
                search="TextEntry",
                property_map=dict(pv_name="pv"),
            ),
            table_formatter_cls=PVWidgetFormatter.from_template(
                template,
                search="Table",
                property_map=dict(pv_name="pv"),
            ),
            action_formatter_cls=ActionWidgetFormatter.from_template(
                template,
                search="WritePV",
                property_map=dict(text="label", pv_name="pv", value="value"),
            ),
            sub_screen_formatter_cls=SubScreenWidgetFormatter.from_template(
                template,
                search="OpenDisplay",
                property_map=dict(file="file_name", text="label", macros="macros"),
            ),
        )
        # MAKE_WIDGETS DOCS REF: Define screen and group widgets
        screen_title_cls = LabelWidgetFormatter.from_template(
            template,
            search="Title",
            property_map=dict(text="text"),
        )
        group_title_cls = LabelWidgetFormatter.from_template(
            template,
            search="Group",
            property_map=dict(name="text"),
        )

        def create_group_object_formatter(
            bounds: Bounds, title: str
        ) -> List[WidgetFormatter[str]]:
            return [
                group_title_cls(
                    Bounds(bounds.x, bounds.y, bounds.w, bounds.h), f"{title}"
                )
            ]

        def create_screen_title_formatter(
            bounds: Bounds, title: str
        ) -> List[WidgetFormatter[str]]:
            return [
                screen_title_cls(
                    Bounds(0, 0, bounds.w, screen_layout.title_height), title
                )
            ]

        # SCREEN_INI DOCS REF: Construct a screen object
        formatter_factory = ScreenFormatterFactory(
            screen_formatter_cls=GroupFormatter.from_template(
                template,
                search=GroupType.SCREEN,
                sized=with_title(screen_layout.spacing, screen_layout.title_height),
                widget_formatter_hook=create_screen_title_formatter,
            ),
            group_formatter_cls=GroupFormatter.from_template(
                template,
                search=GroupType.GROUP,
                sized=with_title(
                    screen_layout.spacing, screen_layout.group_label_height
                ),
                widget_formatter_hook=create_group_object_formatter,
            ),
            widget_formatter_factory=widget_formatter_factory,
            prefix=prefix,
            layout=screen_layout,
            base_file_name=path.stem,
        )
        # SCREEN_FORMAT DOCS REF: Format the screen
        title = f"{device.label} - {prefix}"

        screen_formatter, sub_screens = formatter_factory.create_screen_formatter(
            device.children, title
        )

        write_bob(screen_formatter, path)
        for sub_screen_name, sub_screen_formatter in sub_screens:
            sub_screen_path = Path(path.parent / f"{sub_screen_name}{path.suffix}")
            write_bob(sub_screen_formatter, sub_screen_path)


def write_bob(screen_formatter: GroupFormatter, path: Path):
    # SCREEN_WRITE DOCS REF: Generate the screen file
    # The root:'Display' is always the first element in texts
    texts = screen_formatter.format()
    ET = etree.fromstring(etree.tostring(texts[0]))
    for element in texts[:0:-1]:
        ET.insert(ET.index(ET.find("grid_step_y")) + 1, element)
    ET = ET.getroottree()
    ET.write(str(path), pretty_print=True)
